from typing import List, Iterable, Tuple

from spacy.tokens import Span, Doc
from spacy.pipeline import TrainablePipe
from spacy.vocab import Vocab
from spacy.training import Example
from thinc.api import Model
from thinc.types import Ragged, Floats2d


from spacy.pipeline import SpanCategorizer, EntityRecognizer
from spacy import Language

from spacy.pipeline.spancat import DEFAULT_SPANCAT_SINGLELABEL_MODEL
import spacy


from distancer import Distancer
from embedding_utils import EmbeddingUtils
from context_window import ContextWindow
from document import Document
from annotation import Annotation

import math

"""
class EntityRecognizer():

    def __init__(self):
        pass     

    def get_loss(self, examples: Iterable[Example], span_scores: Tuple[Ragged, Floats2d]):
        
        losses =  [self.get_individual_loss(doc, gold) for doc, gold in examples]
        loss_sum = sum(losses)
        gradient = 0.0
        loss = loss_sum / len(losses)
        return (loss, gradient)

    
    def set_annotations(self, docs):
        for doc, gold in docs:
            entities = []
            for entity in gold.get("entities", []):
                start, end, label = entity
                entities.append((start, end, label, "B"))

                for i in range(start + 1, end):
                    entities.append((i, i+1, label, "I"))

                entities[-1] = entities[-1][:3] + ("E",)

            doc.ents = entities

    def predict(self, docs):
    
        verbose = False

        for doc in docs:
            entities = []
            current_label = None
            for token in doc:
                if token.ent_iob_ == "B":
                    if current_label:
                        entities[-1] = entities[-1][:2] + ("E",)
                    entities.append((token.i, token.i + 1, token.ent_type_, "B"))
                    current_label = token.ent_type_
                elif token.ent_iob_ == "I" and current_label == token.ent_type_:
                    continue
                elif token.ent_iob_ == "I" and current_label != token.ent_type_:
                    entities[-1] = entities[-1][:2] + ("E",)
                    entities.append((token.i, token.i + 1, token.ent_type_, "B"))
                    current_label = token.ent_type_
                else:
                    if current_label:
                        if verbose:
                            print("the curent lable is not none")
                        entities[-1] = entities[-1][:2] + ("E",)
                    current_label = None
            if current_label:
                entities[-1] = entities[-1][:2] + ("E",)
            doc.ents = entities
    
    
"""


class EntityRecognizer:
    """
    EntityRecognizer

    A class that houses a SpaCy SpanCategorizer which is repurposed as an entity recognizer.

    spancat_config: Dict[str, any]
        The config dict for the spancat to allow for entity-recognition-like behavior
    key: str
        The key to house the recognized spans under
    span_categorizer: SpanCategorizer
        The span categorizer pipeline object
    nlp: Language
        A SpaCy lanugage set that houses the spancategorizer

    """

    def __init__(self, key: str):
        spancat_config = {
            "spans_key": key,
            "model": DEFAULT_SPANCAT_SINGLELABEL_MODEL,
            "suggester": {"@misc": "spacy.ngram_suggester.v1", "sizes": [1, 2, 3, 4]},
            "negative_weight": 0.8,
            "allow_overlap": True,
            # "add_negative_label": True
        }

        self.key = key

        nlp = spacy.blank("en")
        self.span_categorizer = nlp.add_pipe(
            "spancat_singlelabel", config=spancat_config
        )
        self.span_categorizer.add_label("RELEVANT")
        nlp.initialize()
        self.nlp = nlp

    def __call__(self, document: Doc):
        """
        __call__

        Puts a list of spans representing entities into the SpanSet under the given ER's key

        ------
        PARAMS:

        document: Doc
            The document to annotate
        """

        annotated_doc = self.recognize_entities_in_text(document.text)

        return annotated_doc

    def recognize_entities_in_text(self, text: str):
        """
        recognize_entities_in_text

        Puts spans recognized as mentions in a Doc's span property under the recognizer's span key

        ------
        PARAMS:

        text: str
            The text to return as an annotated SpaCy Doc.
        """
        annotated_doc = self.nlp(text)
        return annotated_doc
